# SimpleFS - Pure Tcl Virtual File System driver
# Copyright (C) 2019 Clif Flynt <clif@noucorp.com>
# Derived in parts from sample VFS in tclvfs distro and 
# sqlite3vfs.tcl -- TclSqlite3 Virtual File System driver
# Copyright (C) 2005 Tiago Dionizio <tngd81@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
#
# History:
#
# Visible interface:
#
#   vfs::simplefs::Mount file local ?args ...? : fd
#       Mount the database in 'file' as directory 'local'.
#       Returns the file descriptor for the new filesystem.
#
#       args:
#           -timeout value      # lock timeout
#           -busy script        # busy lock handler (overrides timeout)
#           -key key            # encryption key
#           -compress level     # compression level (0 = store)
#
#   vfs::simplefs::Unmount fd local
#       Unmounts the filesystem identified by 'fd' mounted in
#       'local' path.
#
#   vfs::simplefs::Runlock fd script ?locktype?
#       Make changes in the filesystem in a transaction which
#       may improve performance when doing multiple operations.
#       The 'locktype' can be DEFERRED or IMMEDIATE (values that
#       make sense to use here) see the sqlite3 documentation
#       for more information on transaction types.
#       The script 'script' will be run and the transaction
#       will be closed when it finishes execution. If it raises
#       an error, the transaction will be rolled back and
#       invalidating any changes made during execution.

package require vfs 1.3             ;# virtual filesystem
catch {package require zlib }    ;# for compression
package require Memchan             ;# memchan is used a bunch

package provide simplevfs 1.0
package provide vfs::simplefs 1.0


namespace eval vfs::simplefs {

    # place all internal procs and variables in here
    namespace eval _ {
        variable mountPt

        # database names are numbers
        variable seq 0

        # posix error to raise when an error occours
        variable perror

        # transaction lock count
        variable lock

        # filesystem options
        variable option

        array set lock {}
        array set perror {}
        array set option {}

    }
}

# Mount the simpleFS 
# fsfile	File to mount (may be self)
# local		Local mount point.

proc vfs::simplefs::Mount {fsfile local args} {
    variable fs
    variable mountPt

    # follow links
    set fsfile [::vfs::filesystem fullynormalize $fsfile]
# puts "---[namespace current] --    set mountPt([namespace current]::$fsfile) $local"; flush stdout

    set mountPt([namespace current]::$fsfile) $local


    set fs [simpleFS create $fsfile]
    $fs mount $fsfile

    # $fs pprint

    # puts "MOUNT: $fs"
    ::vfs::filesystem mount $local [list [namespace current]::handler $fs]
    ::vfs::RegisterMount $local [list [namespace current]::Unmount $fs]
vfs::filesystem  internalerror  report  ;  proc report {} { puts stderr "VFS_ERR: $::errorInfo"} 
    return $fs
}

# Unmount the sqlite3 database file

proc vfs::simplefs::Unmount {fs local} {
    variable mountPt
    # puts "Unmount FS: $fs LOC: $local "
    ::vfs::filesystem unmount $local
    $fs destroy
    unset mountPt($fs)
}

proc vfs::simplefs::cvrt2Actual {fs actual FSpath} {
  variable mountPt
   # puts "** CVRT2ACTUAL FS: $fs -- MTPT: $mountPt($fs) ACTUAL: $actual FSPATH: $FSpath"
  if {$actual eq $mountPt($fs)} {
    set rtn "$mountPt($fs)/$FSpath"
  } else {
    set p1 [string first $mountPt($fs) $actual]
    if {$p1 == 0} {
      # Actual was the mount point
      if {[string first $mountPt($fs) $FSpath] < 0} {
        set rtn $mountPt($fs)/$FSpath
      } else {
        set rtn $FSpath
      }
    } elseif {$p1 < 0} {
       set p2 [string first $actual $FSpath]
       if {$p2 > 0} {
         set rtn [string range $FSpath [expr {$p2 + [string length $p2]-1}] end]
       } else {
         set rtn $FSpath
       }
    } else {
      # p1 > 0
      set p2 [string first $mountPt($fs) $actual]
      if {$p2 > 0} {
# puts "MAPPING ACTUAL"
	 set head [lindex [split [string trim $FSpath /] /] 0]
	 set p1 [string first /$head $actual]
#         set rtn [file dirname $actual]/$FSpath
         incr p1 -1
         set rtn [string range $actual 0 $p1]/$FSpath
      } else {
        # punt
        set rtn $FSpath
      }
    }
  }

  if {$rtn eq ""} {
    set rtn "/"
  }
#  puts "CVRT2ACTUAL RTN: '$rtn'"
  return $rtn
}

proc vfs::simplefs::cvrt2FS {fs actual} {
  variable mountPt
  # puts "** CVRT2FS FS: $fs -- MTPT: [array get mountPt] actual: $actual"
  if {$actual eq $mountPt($fs)} {
    set rtn "/"
  } else {
    set p1 [string first $mountPt($fs)/ $actual]
    if {$p1 >= 0} {
      # The actual is within the FS filesystem
      set rtn [string range $actual [expr {$p1 + [string length $mountPt($fs)]}] end ]
    } else {
      set p1 [string first /$mountPt($fs) $actual]
      if {$p1 >= 0} {
        # The Actual is above
        set rtn [string range $actual \
	    [expr {1+$p1 + [string length $mountPt($fs)]}] end ]
      } else {
        # puts "WHA: NO $mountPt($fs) in $actual?"
	set files [$fs listing]
	set el [lsearch -inline $files */${actual}/*]
	set p1 [lsearch $actual $el]
	set path [string range $el 0 [expr {1+$p1+[string length $actual]}]]
	# puts "EL: $el PATH: $path FILES $files"
        set rtn $path
      }
    }
  }

  if {$rtn eq ""} {
    set rtn "/"
  }
  #  puts "CVRT2FS: RTN: '$rtn'"
  return $rtn
}


## vfshandler subcmd root relative actualpath args...
 #     The first argument specifies the operation to perform on behalf of
 #     the filesystem code in the tcl core, the remainder specify the file
 #     path on which to operate, in different forms, and parts, and any
 #     additional arguments which may be required to carry out the action.
 #
 #     To demonstrate the treatment of a path by the generic layer we use
 #     "C:/foo/bar/mount.zip/xxx/yyy" as an example and additionally assume
 #     that the following conditions are true:
 #
 #        1. "mount.zip" is a zip archive which has been mounted on top of
 #           itself,
 #        2. said zip archive contains a file with path "xxx/yyy",
 #        3. the current working directory of the application is inside of
 #           directory "xxx",
 #        4. and the command executed is file exists yyy.
 #
 #     The file separator between root and relative is omitted. Most
 #     filesystem operations need only the relative argument for their
 #     correct operation, but some actually require the other parts of the
 #     path.
 #
 #     subcmd
 #         This argument of the handler can be one of the following access,
 #         createdirectory, deletefile, fileattributes, matchindirectory,
 #         open, removedirectory, stat, or utime.
 #
 #         The generic layer expects that the subcommands of a handler
 #         signal error conditions by calling vfs::filesystem posixerror
 #         with the appropriate posix error code instead of throwing a tcl
 #         error. If the latter is done nevertheless it will be treated as
 #         an unknown posix error.
 #
 #         There are three exceptions to the rule above: If any of open
 #         (when an interpreter is given), matchindirectory, and
 #         fileattributes (for a set or get operation only) throw a tcl
 #         error, this error will be passed up to the caller of the
 #         filesystem command which invoked the handler. Note that this
 #         does not preclude the ability of these subcommands to use the
 #         command vfs::filesystem posixerror to report more regular
 #         filesystem errors.
 #
 #     root
 #         Part of the specification of the path to operate upon. It
 #         contains the part of the path which lies outside this
 #         filesystem's mount point. For example outlined above its
 #         value will be "C:/foo/bar/mount.zip".
 #
 #     relative
 #         Part of the specification of the path to operate upon. It
 #         contains the part of the path which lies inside this
 #         filesystem. For example outlined above its value will be
 #         "xxx/yyy".
 #
 #     actualpath
 #         Part of the specification of the path to operate upon. It
 #         contains the original (unnormalized) name of the path which
 #         was used in the current command wherever it originated (in
 #         Tcl or C). For example outlined above its value will be "yyy".
 ##

proc vfs::simplefs::handler {fs subcmd args} {
#      puts "*** handler :: $fs subcmd $subcmd -- $args"

    # Choose appropriate locking mechanism.
    # Hopefully, this will help with concurrency issues; when aquiring
    # reserved locks when there is already a reader, will cause the
    # busy handler to be invoqued and wait for a while to try to get
    # write lock on the database. Can be looked at as a potencial
    # blocking operation. This is intended to avoid as many locking
    # problems as possible, without failing to execute later sql statements.
    # Note that these functions are not recursive, so there should be
    # no problems here, unless user calls RunLock proc with an inadequate
    # lock (such as shared lock when a reserved one is needed and there is
    # another writer doing some work).
    set lock DEFERRED ;# reading only

#  puts "SUBCMD: $subcmd fs: $fs"
    set ::errorInfo "NO-ERR"
    if {[catch {eval [list $subcmd $fs] $args} res]} {
         # puts "CAUGHT ERROR: in SUB: '$subcmd' with ARGS: '$args'\n$::errorInfo"
         # puts "$::errorInfo"; flush stdout
	return -code -1 $res
    } else {
# puts "HANDLER RTN: $res"; flush stdout
        return -code 0 $res 
    }
}

## vfshandler access root relative actualpath mode
 #     Signal a posix error if the specified access mode (an integer
 #     number) is not compatible with the file or directory described
 #     by the path. The generic layer will ignore any non-empty return
 #     value.
 #
 #     The command vfs::accessMode (see section HANDLER ENVIRONMENT)
 #     can be used to convert the integer mode into an easier to check
 #     string value.
 ##

#Failure if pass actual to cvrt2fs
# Condition after cd and glob *
#ACCESS: fs ::vfs::simplefs::/clif/TCL_STUFF/ObfusKit/xx 
#root /clif/TCL_STUFF/ObfusKit/foo 
#relative d0/tablelist 
#actualpath tablelist mode 4
#
#** CVRT2FS FS: ::vfs::simplefs::/clif/TCL_STUFF/ObfusKit/xx -- 
#MTPT: ::vfs::simplefs::/clif/TCL_STUFF/ObfusKit/obfuskit simpliskit 
#::vfs::simplefs::/clif/TCL_STUFF/ObfusKit/xx foo 
#actual: tablelist

#Failure if pass relative to cvrt2fs
# ACCESS: fs ::vfs::simplefs::/clif/TCL_STUFF/ObfusKit/obfuskit 
# root /clif/TCL_STUFF/ObfusKit/simpliskit 
# relative simpleVFS.vfs/main.tcl 
# actualpath simpliskit/simpleVFS.vfs/main.tcl mode 0
#
# ** CVRT2FS FS: ::vfs::simplefs::/clif/TCL_STUFF/ObfusKit/obfuskit -- 
# MTPT: ::vfs::simplefs::/clif/TCL_STUFF/ObfusKit/obfuskit simpliskit 
# actual: simpleVFS.vfs/main.tcl
# CVRT2FS: RTN: '/'

proc vfs::simplefs::access {fs root relative actualpath mode} {
# puts "ACCESS: fs $fs root $root relative $relative actualpath $actualpath mode $mode "; flush stdout
  variable mountPt

catch {
#  set path [cvrt2FS $fs $actualpath] ;# Fails if cd into filesystem.
#  set path [cvrt2FS $fs $relative] ;# Just fails
   set path [cvrt2FS $fs [::vfs::filesystem fullynormalize $actualpath]]

} err
# puts "ERRXX: $err"
# puts "FS ACCESS PATH: $path"

set ::errorInfo ""

  if {[$fs access $path $mode] == 0} {
#   puts "FS ACCESS FAIL- MODE: $mode ERR $::errorInfo"; flush stdout
     if {($mode eq "R") || ($mode == 4) || ($mode == 0)} {
#   puts "FS ACCESS FOLDER CHECK"; flush stdout
      #check for a directory
      if {[$fs access $path/* $mode] == 1} {
# puts "FS ACCESS OK"
        return 0
      }
    }
#  puts "FS ACCESS RETURN FAIL"; flush stdout
    set ::errorInfo "No File $path"
    return -code -1 -1
  }

# puts "FS ACCESS OK 2"
  return 0
}

## vfshandler deletefile root relative actualpath
 #     Delete the given file.
 ##

proc vfs::simplefs::deletefile {fs root relative actualpath} {
# puts "deletefile fs $fs root $root relative $relative actualpath $actualpath "

  $fs delFile /$relative
#  $fs delFile /[join [lrange [file split $actualpath] 1 end] /]
}

## vfshandler fileattributes root relative actualpath ?index? ?value?
 #     The command has to return a list containing the names of all
 #     acceptable attributes, if neither index nor value were specified.
 #
 #     The command has to return the value of the index'th attribute if
 #     the index is specified, but not the value. The attributes are
 #     counted in the same order as their names appear in the list
 #     returned by a call where neither index nor value were specified.
 #     The first attribute is has the index 0.
 #
 #     The command has to set the value of the index'th attribute to
 #     value if both index and value were specified for the call.
 ##

proc vfs::simplefs::fileattributes {db root relative actualpath {index {}} {value {}} args} {
# puts "vfs::simplefs::fileattributes db $db root $root relative $relative actualpath $actualpath index $index value $value args $args "
    switch -- [llength $args] {
        0 {
            # list strings
            return [::vfs::listAttributes]
        }
        1 {
            # get value
            set index [lindex $args 0]
            return [::vfs::attributesGet $root $relative $index]
        }
        2 {
            # set value
            #if {$::mode($db) == "readonly"} {
            #    vfs::filesystem posixerror $::vfs::posix(EROFS)
            #}
            #set index [lindex $args 0]
            #set val [lindex $args 1]
            return [::vfs::attributesSet $root $relative $index $val]
        }
    }
}

## vfshandler matchindirectory root relative actualpath pattern types
 #     Return the list of files or directories in the given path which
 #     match the glob pattern and are compatible with the specified
 #     list of types. The specified path is always the name of an
 #     existing directory.
 #
 #     Note: As Tcl generates requests for directory-only matches from
 #     the filesystems involved when performing any type of recursive
 #     globbing this subcommand absolutely has to handle such (and
 #     file-only) requests correctly or bad things (TM) will happen.
 #
 #     The commands vfs::matchDirectories and vfs::matchFiles (see
 #     section HANDLER ENVIRONMENT) can aid the implementation greatly
 #     in this task.
 ##

proc vfs::simplefs::UNUSEDgetdirs {fs ptrn args} {
  puts "GETDIRS: vfs::simplefs::getdirs fs: $fs ptrn $ptrn args $args" 
    return /
}

proc vfs::simplefs::UNUSEDgetfiles {fs relative pattern} {
 puts "GETFILES vfs::simplefs::getfiles fs $fs relative $relative pattern $pattern "

    set files [$fs doDir $pattern]
    

    return $files
}

proc vfs::simplefs::useElement {fs types nm} {
# puts "useelement: TYPES: $types NM: $nm"
catch {
  set rtn 0
  array set lkup {directory 4 file 16}

  if {([lsearch $types 0] >= 0) || ([lsearch $types 20] >= 0)} {
    # puts "USEEL RET 1"
    set rtn 1
  } else {
    set statDct [$fs stat $nm] 
# puts "STATDCT: $statDct"
    if {[dict get $statDct type] eq "deleted"} {
      # puts "USEEL RET 0"
      set rtn 0
    } else {
      set num $lkup([dict get $statDct type])
      foreach typ $types {
        if {$num eq $typ} {
          set rtn 1
        }
      }
    }
  }
} xx

# puts "USEEL RTN: $rtn -- $xx"
  return $rtn
}

proc vfs::simplefs::matchindirectory {fs root relative actualpath pattern types} {
 #   puts "---matchindirectory FS $fs \nROOT $root REL $relative ACT $actualpath PTRN $pattern TYPE $types \nPWD [pwd]"; flush stdout

# Type definitions from generic/tcl.h
#define TCL_GLOB_TYPE_BLOCK             (1<<0)   1
#define TCL_GLOB_TYPE_CHAR              (1<<1)   2
#define TCL_GLOB_TYPE_DIR               (1<<2)   4
#define TCL_GLOB_TYPE_PIPE              (1<<3)   8
#define TCL_GLOB_TYPE_FILE              (1<<4)  16
#define TCL_GLOB_TYPE_LINK              (1<<5)  32
#define TCL_GLOB_TYPE_SOCK              (1<<6)  64
#define TCL_GLOB_TYPE_MOUNT             (1<<7) 128

    set actual_match {}

    set rootedRelative [cvrt2FS $fs [::vfs::filesystem fullynormalize $actualpath]]
      variable mountPt
# puts "ROOTEDRELATIVE: $rootedRelative FROM $actualpath FROM [::vfs::filesystem fullynormalize $actualpath]"
    if {$pattern eq ""} {
        # match single file
# puts "DIR 1"
        set match [$fs doDir $rootedRelative $types]
	set newmatch {}
	foreach m $match {
	  lappend newmatch [string trimleft $m /]
	}
	set match $newmatch
	set normalize 0
    } else {
        if {$rootedRelative eq "/"} {set rootedRelative ""}
        set match [$fs doDir $rootedRelative/$pattern $types]
# puts "DIR 2 Match: $match"
	set newmatch {}
	foreach m $match {
	  lappend newmatch [string trimleft $m /]
	}
	set match $newmatch
	set normalize 1
    }

# puts "MATCH1: $match - "
# puts "ROOT: $root ACT: $actualpath"
      set ::errorInfo "NOERR"
catch {
      foreach mm [lsort $match] {
        if {[useElement $fs $types /$mm]} {
    	  lappend actual_match [cvrt2Actual $fs $actualpath $mm]
  	}
      }
} xx
# puts "XX: $xx"

# ACTUALMATCH2: /clif/TCL_STUFF/ObfusKit/foo/d0/d1 /clif/TCL_STUFF/ObfusKit/foo/d0/f2-1.txt

# puts "ACTUAL_MATCH: $actual_match ACTPATH: $actualpath"
  
  set new {}
  foreach n $actual_match {
#    lappend new [string trimleft $n /]
     if {$n ne $actualpath} {
       set n [string map [list $actualpath ""] $n]
       lappend new [file join $actualpath [string trimleft $n /]]
     } else {
       lappend new $n
     }
  }
  set actual_match $new

#  puts "MATCHINDIR RETURN: $actual_match"; flush stdout;
    return $actual_match
}


## vfshandler open root relative actualpath mode permissions
 #     Either returns a list describing the successfully opened file,
 #     or throws an error describing how the operation failed.
 #
 #     The list returned upon success contains at least one and at
 #     most two elements. The first, obligatory, element is always
 #     the handle of the channel which was created to allow access
 #     to the contents of the file.
 #
 #     If specified the second element will be interpreted as a
 #     callback, i.e. a command prefix. This prefix will always be
 #     executed as is, i.e. without additional arguments. Any
 #     required arguments have to be returned as part of the result
 #     of the call to open.
 #
 #     If present the specified callback will be evaluated just before
 #     the channel is closed by the generic filesystem layer. The
 #     callback itself must not call close.
 #
 #     The channel however is live enough to allow seek and read
 #     operations. In addition all available data will have been
 #     flushed into it already. This means, for example, that the
 #     callback can seek to the beginning of the said channel, read
 #     its contents and then store the gathered data elsewhere. In
 #     other words, this callback is not only crucial to the cleanup
 #     of any resources associated with an opened file, but also for
 #     the ability to implement a filesystem which can be written to.
 #
 #     Under normal circumstances return code and any errors thrown
 #     by the callback itself are ignored. In that case errors have
 #     to be signaled asychronously, for example by calling bgerror.
 #     However if, through a call of the subcommand internalerror,
 #     an error handling script has been specified for the file
 #     system, all errors thrown here will be passed to that script
 #     for further action.
 #
 #     mode
 #         can be any of r, w, a, w+, or a+.
 #
 #     permissions
 #         determines the native mode the openend file is created
 #         with. Relevant only of the open mode actually requests
 #         the creation of a non-existing file, i.e. is not r.
 ##

proc vfs::simplefs::writefile {fs relative fd} {
# puts "--writefile FS: $fs RELative: $relative FD: $fd"
    if {$type eq "file"} {
        variable option

        fconfigure $fd -translation binary
        seek $fd 0
        set contents [read $fd]
        set size [string bytelength $contents]
        set mtime [clock seconds]

        if {$compress} {
            # compress contents
            catch {
                set ncontents [zlib compress $contents $compress]
                # replace only if size is smaller
                if {[string bytelength $ncontents] < $size} {
                    set contents $ncontents
                }
            }
        }

        set sql {UPDATE files SET size = $size, contents = $contents, mtime = $mtime WHERE id = $id}

        if {[catch {$db eval $sql}]} {
            catch {end $db 1}
            throw $db EIO
        } else {
            end $db
        }
    }
}

proc vfs::simplefs::close {args} {
  variable opens
  variable mountPt

#   puts "CLOSE : ARGS: $args"
  lassign $args fs fd rel
  # puts "FS: $fs FD: $fd REL: $rel"

  if {[info exists opens($fd)]} {
    seek $fd 0 start
    catch {
      set d [read $fd]
    }
  # puts "READ: ERR: $::errorInfo LEN: [string length $d] MTPT: $mountPt($fs)"

    if {[string first $mountPt($fs) $rel] == 1} {
      set rel [string range $rel [expr {[string length $mountPt($fs)]+2}] end]
    }

    # puts "CLOSE    $fs writeFile $rel 'obs_shrink $d'"
#    $fs writeFile $rel [obs_shrink $d]
    $fs writeFile $rel $d
  # puts "WRITEFILE ERR2: $::errorInfo FS: $fs LEN D: [string length $d] Len Shrink: [string length $d]"
  }
}

proc vfs::simplefs::open {fs root relative actualpath mode permissions} {
  variable opens
  # $fs pprint
  # puts "VFS_OPEN vfs::simplefs::open fs $fs root $root relative $relative actualpath $actualpath mode $mode permissions $permissions "
    # return a list of two elements:
    # 1. first element is the Tcl channel name which has been opened
    # 2. second element (optional) is a command to evaluate when
    #  the channel is closed.
    
    set fd [::vfs::memchan]

    fconfigure $fd -encoding binary
    fconfigure $fd -translation binary

    set relative "/$relative"

    switch $mode {
      {} -
      "r" -
      "r+" {
        if {[catch {set  dd [$fs readFile $relative ]}]} {
           # puts "ERR: OPENING: $relative $::errorInfo "
           error "Failed to open $relative" "Failed to open $relative"
       } else {
         #  puts "OPEN OK: $relative"
          set ::errorInfo ""
      }

      puts $fd $dd

      seek $fd 0 start
      }
      "w" -
      "w+" {
        set opens($fd) $relative
      }
      "a" -
      "a+" {
        error "Append not supported."
      }
      default {
        error "INVALID MODE: $mode"
      }
    }
    # puts "open to RETURN: $fd [namespace current]::close $fd $relative"
    return [list $fd "[namespace current]::close $fs $fd $relative"]
}

## vfshandler removedirectory root relative actualpath recursive
 #     Delete the given directory. Argument recursive is a boolean.
 #     If the specified value is true then even if the directory is
 #     non-empty, an attempt has to be made to recursively delete it
 #     and its contents. If the spcified value is false and the
 #     directory is non-empty, a posix error (EEXIST) has to be thrown.
 ##

proc vfs::simplefs::removedirectory {fs root relative actualpath recursive} {
#  puts "removedirectory fs $fs root $root relative $relative actualpath $actualpath recursive $recursive "
    # get dir name

    set rootedRelative [cvrt2FS $fs [::vfs::filesystem fullynormalize $actualpath]]
# puts "ROOTED: $rootedRelative"

    if {$recursive} {
      set all [$fs listing]
      set files [lsearch -inline -all $all $rootedRelative/*]

      foreach f $files {
        switch [$fs fileOrDir $f] {
	  "file" {
	    $fs delFile $f
	  }
	  "directory" {
	    removedirectory $fs $root [cvrt2Actual $actual] [cvrt2Actual $actual]
	  }
	}
      }
      $fs compressFiles
    } else {
      # Empty folders don't exist in simpleFS
      return
    }
}

## vfshandler stat root relative actualpath
 #     The result has to be a list of keys and values, in a format
 #     acceptable to the builtin command array set. It describes the
 #     contents of a stat structure. The order of the keys in the
 #     list is not important.
 #
 #     Given this the subcommand should use something like
 #
 #         return [list dev 0 type file mtime 1234 ...].
 #
 #     as the last command of its implementation.
 #
 #     The following keys and their values have to be supplied by the
 #     filesystem:
 #
 #     dev
 #         A long integer number, the device number of the path stat
 #         was called for.
 #
 #     ino
 #         A long integer number, the inode number of the path stat
 #         was called for. Each path handled by the filesystem should
 #         be uniquely identified by the combination of device and
 #         inode number. Violating this principle will cause
 #         higher-level algorithms which(have to) keep track of device
 #         and inode information to fail in all manners possible.
 #
 #         An example of such an algorithm would be a directory walker
 #         using device/inode information to keep itself out of
 #         infinite loops generated through symbolic links. Returning
 #         non-unique device/inode information will most likely cause
 #         such a walker to skip over paths under the wrong assumption
 #         of having them seen already.
 #
 #     mode
 #         An integer number, the access mode of the path. It is this
 #         mode which is checked by the subcommand access.
 #
 #     nlink
 #         A long integer number, the number of hard links to the
 #         specified path.
 #
 #     uid
 #         A long integer number, the id of the user owning the
 #         virtual path.
 #
 #     gid
 #         A long integer number, the id of the user group the
 #         virtual path belongs to.
 #
 #     size
 #         A long integer number, the true size of the virtual path,
 #         in bytes.
 #
 #     atime
 #         A long integer number, the time of the latest access to
 #         the path, in seconds since the epoch. Convertible into a
 #         readable date/time by the command clock format.
 #
 #     mtime
 #         A long integer number, the time of the latest modification
 #         of the path, in seconds since the epoch. Convertible into a
 #         readable date/time by the command clock format.
 #
 #     ctime
 #         A long integer number, the time of the path was created, in
 #         seconds since the epoch. Convertible into a readable
 #         date/time by the command clock format.
 #
 #     type
 #         A string, either directory, or file, describing the type of
 #         the given path.
 ##

proc vfs::simplefs::stat {fs root relative actualpath} {
#  puts "stat fs $fs root $root relative $relative actualpath $actualpath "

  set d1 [$fs stat /$relative]

#  puts "STAT RTN: [list dev 0 ino 0 mode 0x0777 nlink 1 uid 0 gid 0 \
#      size [dict get $d1 size] atime [clock seconds] mtime [clock seconds] \
#      ctime [clock seconds] type [dict get $d1 type] blocks 0 ]"
  # NOT 0x777 - OCTAL 777 - decimal 511.
  
  if {[dict get $d1 type] eq "file"} {
    # copied from a file with 777 perms
    set mode 33279
  } else {
    # Copied from /tmp
    set mode 17407
  }
  return [list dev 0 ino 0 mode $mode nlink 1 uid 0 gid 0 \
      size [dict get $d1 size] atime [clock seconds] mtime [clock seconds] \
      ctime [clock seconds] type [dict get $d1 type] blocks 0 ]
}

proc vfs::simplefs::utime {fs args} {
#  puts "utime fs $fs root $root relative $relative actualpath $actualpath "
#  puts "utime fs $fs args: $args"

#  puts "utime Rtn:   [clock seconds]"
  return [clock seconds]
}




## HANDLER ENVIRONMENT
 #
 # The implementation of a filesystem handler can rely on the
 # existence of the following utility commands:
 #
 # vfs::accessMode mode
 #     This commands converts an access mode given as integer
 #     into a string, one of F, X, W, XW, R, RX, and RW.
 #
 # vfs::matchDirectories types
 #     Checks if the glob types specification ask for the inclusion
 #     of directories. Returns a boolean result. true is returned
 #     if types does ask for directories, else false.
 #
 # vfs::matchFiles types
 #     Checks if the glob types specification ask for the inclusion
 #     of files. Returns a boolean result. true is returned if types
 #     does ask for directories, else false.
 #
 # vfs::matchCorrectTypes types filelist ?inDir?
 #     Returns that subset of the filelist which are compatible with
 #     the types given. The elements of filelist are either absolute
 #     paths, or names of files in the directory indir. The latter
 #     interpretation is taken if and only if the argument indir is
 #     specified.
 ##


